# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

from airflow.providers.amazon.aws.hooks.eks import ClusterStates, NodegroupStates
from airflow.providers.amazon.aws.operators.eks import (
    EksCreateClusterOperator,
    EksCreateNodegroupOperator,
    EksDeleteClusterOperator,
    EksDeleteNodegroupOperator,
    EksPodOperator,
)
from airflow.providers.amazon.aws.sensors.eks import EksClusterStateSensor, EksNodegroupStateSensor
from airflow.providers.common.compat.sdk import DAG, chain

from system.amazon.aws.utils import SystemTestContextBuilder

sys_test_context_task = SystemTestContextBuilder().build()

DAG_ID = "example_eks_templated"

# Example Jinja Template format, substitute your values:
# {
#     "cluster_name": "templated-cluster",
#     "cluster_role_arn": "arn:aws:iam::123456789012:role/role_name",
#     "resources_vpc_config": {
#         "subnetIds": ["subnet-12345ab", "subnet-67890cd"],
#         "endpointPublicAccess": true,
#         "endpointPrivateAccess": false
#     },
#     "nodegroup_name": "templated-nodegroup",
#     "nodegroup_subnets": "['subnet-12345ab', 'subnet-67890cd']",
#     "nodegroup_role_arn": "arn:aws:iam::123456789012:role/role_name"
# }

with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    tags=["example", "templated"],
    catchup=False,
    # render_template_as_native_obj=True is what converts the Jinja to Python objects, instead of a string.
    render_template_as_native_obj=True,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context["ENV_ID"]

    CLUSTER_NAME = "{{ dag_run.conf['cluster_name'] }}"
    NODEGROUP_NAME = "{{ dag_run.conf['nodegroup_name'] }}"

    # Create an Amazon EKS Cluster control plane without attaching a compute service.
    create_cluster = EksCreateClusterOperator(
        task_id="create_eks_cluster",
        cluster_name=CLUSTER_NAME,
        compute=None,
        cluster_role_arn="{{ dag_run.conf['cluster_role_arn'] }}",
        # This only works with render_template_as_native_obj flag (this dag has it set)
        resources_vpc_config="{{ dag_run.conf['resources_vpc_config'] }}",  # type: ignore[arg-type]
    )

    await_create_cluster = EksClusterStateSensor(
        task_id="wait_for_create_cluster",
        cluster_name=CLUSTER_NAME,
        target_state=ClusterStates.ACTIVE,
    )

    create_nodegroup = EksCreateNodegroupOperator(
        task_id="create_eks_nodegroup",
        cluster_name=CLUSTER_NAME,
        nodegroup_name=NODEGROUP_NAME,
        nodegroup_subnets="{{ dag_run.conf['nodegroup_subnets'] }}",
        nodegroup_role_arn="{{ dag_run.conf['nodegroup_role_arn'] }}",
    )

    await_create_nodegroup = EksNodegroupStateSensor(
        task_id="wait_for_create_nodegroup",
        cluster_name=CLUSTER_NAME,
        nodegroup_name=NODEGROUP_NAME,
        target_state=NodegroupStates.ACTIVE,
    )

    start_pod = EksPodOperator(
        task_id="run_pod",
        cluster_name=CLUSTER_NAME,
        pod_name="run_pod",
        image="amazon/aws-cli:latest",
        cmds=["sh", "-c", "ls"],
        labels={"demo": "hello_world"},
        get_logs=True,
        # Delete the pod when it reaches its final state, or the execution is interrupted.
        on_finish_action="delete_pod",
    )

    delete_nodegroup = EksDeleteNodegroupOperator(
        task_id="delete_eks_nodegroup",
        cluster_name=CLUSTER_NAME,
        nodegroup_name=NODEGROUP_NAME,
    )

    await_delete_nodegroup = EksNodegroupStateSensor(
        task_id="wait_for_delete_nodegroup",
        cluster_name=CLUSTER_NAME,
        nodegroup_name=NODEGROUP_NAME,
        target_state=NodegroupStates.NONEXISTENT,
    )

    delete_cluster = EksDeleteClusterOperator(
        task_id="delete_eks_cluster",
        cluster_name=CLUSTER_NAME,
    )

    await_delete_cluster = EksClusterStateSensor(
        task_id="wait_for_delete_cluster",
        cluster_name=CLUSTER_NAME,
        target_state=ClusterStates.NONEXISTENT,
    )

    chain(
        # TEST SETUP
        test_context,
        # TEST BODY
        create_cluster,
        await_create_cluster,
        create_nodegroup,
        await_create_nodegroup,
        start_pod,
        # TEST TEARDOWN
        delete_nodegroup,
        await_delete_nodegroup,
        delete_cluster,
        await_delete_cluster,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
from moto import mock_aws

from airflow.providers.amazon.aws.hooks.sqs import SqsHook

QUEUE_URL = "https://sqs.region.amazonaws.com/123456789/test-queue"
MESSAGE_BODY = "test message"

MESSAGE_ID_KEY = "MessageId"


class TestSqsHook:
    @mock_aws
    def test_get_conn(self):
        hook = SqsHook(aws_conn_id="aws_default")
        assert hook.get_conn() is not None


@pytest.mark.asyncio
class TestAsyncSqsHook:
    """The mock_aws decorator uses `moto` which does not currently support async SQS so we mock it manually."""

    @pytest.fixture
    def hook(self):
        return SqsHook(aws_conn_id="aws_default")

    @pytest.fixture
    def mock_async_client(self):
        mock_client = mock.AsyncMock()
        mock_client.send_message.return_value = {MESSAGE_ID_KEY: "test-message-id"}
        return mock_client

    @pytest.fixture
    def mock_get_async_conn(self, mock_async_client):
        with mock.patch.object(SqsHook, "get_async_conn") as mocked_conn:
            mocked_conn.return_value = mock_async_client
            mocked_conn.return_value.__aenter__.return_value = mock_async_client
            yield mocked_conn

    async def test_get_async_conn(self, hook, mock_get_async_conn, mock_async_client):
        # Test context manager access
        async with await hook.get_async_conn() as async_conn:
            assert async_conn is mock_async_client

        # Test direct access
        async_conn = await hook.get_async_conn()
        assert async_conn is mock_async_client

    async def test_asend_message(self, hook, mock_get_async_conn, mock_async_client):
        response = await hook.asend_message(queue_url=QUEUE_URL, message_body=MESSAGE_BODY)

        assert MESSAGE_ID_KEY in response



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import enum
import gzip
import io
import warnings
from collections import namedtuple
from collections.abc import Iterable, Mapping, Sequence
from typing import TYPE_CHECKING, Any, Literal, cast

from airflow.exceptions import AirflowException, AirflowProviderDeprecationWarning
from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.common.compat.sdk import BaseHook, BaseOperator

if TYPE_CHECKING:
    import pandas as pd
    import polars as pl

    from airflow.providers.common.sql.hooks.sql import DbApiHook
    from airflow.utils.context import Context


class FILE_FORMAT(enum.Enum):
    """Possible file formats."""

    CSV = enum.auto()
    JSON = enum.auto()
    PARQUET = enum.auto()


FileOptions = namedtuple("FileOptions", ["mode", "suffix", "function"])

FILE_OPTIONS_MAP = {
    FILE_FORMAT.CSV: FileOptions("r+", ".csv", "to_csv"),
    FILE_FORMAT.JSON: FileOptions("r+", ".json", "to_json"),
    FILE_FORMAT.PARQUET: FileOptions("rb+", ".parquet", "to_parquet"),
}


class SqlToS3Operator(BaseOperator):
    """
    Saves data from a specific SQL query into a file in S3.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:SqlToS3Operator`

    :param query: the sql query to be executed. If you want to execute a file, place the absolute path of it,
        ending with .sql extension. (templated)
    :param s3_bucket: bucket where the data will be stored. (templated)
    :param s3_key: desired key for the file. It includes the name of the file. (templated)
    :param replace: whether or not to replace the file in S3 if it previously existed
    :param sql_conn_id: reference to a specific database.
    :param sql_hook_params: Extra config params to be passed to the underlying hook.
        Should match the desired hook constructor params.
    :param parameters: (optional) the parameters to render the SQL query with.
    :param read_kwargs: arguments to include in DataFrame when reading from SQL (supports both pandas and polars).
    :param df_type: the type of DataFrame to use ('pandas' or 'polars'). Defaults to 'pandas'.
    :param aws_conn_id: reference to a specific S3 connection
    :param verify: Whether or not to verify SSL certificates for S3 connection.
        By default SSL certificates are verified.
        You can provide the following values:

        - ``False``: do not validate SSL certificates. SSL will still be used
                (unless use_ssl is False), but SSL certificates will not be verified.
        - ``path/to/cert/bundle.pem``: A filename of the CA cert bundle to uses.
                You can specify this argument if you want to use a different
                CA cert bundle than the one used by botocore.
    :param file_format: the destination file format, only string 'csv', 'json' or 'parquet' is accepted.
    :param max_rows_per_file: (optional) argument to set destination file number of rows limit, if source data
        is larger than that, it will be dispatched into multiple files.
        Will be ignored if ``groupby_kwargs`` argument is specified.
    :param df_kwargs: arguments to include in DataFrame ``.to_parquet()``, ``.to_json()`` or ``.to_csv()``.
    :param groupby_kwargs: argument to include in DataFrame ``groupby()``.
    """

    template_fields: Sequence[str] = (
        "s3_bucket",
        "s3_key",
        "query",
        "sql_conn_id",
    )
    template_ext: Sequence[str] = (".sql",)
    template_fields_renderers = {
        "query": "sql",
        "df_kwargs": "json",
        "pd_kwargs": "json",
        "read_kwargs": "json",
    }

    def __init__(
        self,
        *,
        query: str,
        s3_bucket: str,
        s3_key: str,
        sql_conn_id: str,
        sql_hook_params: dict | None = None,
        parameters: None | Mapping[str, Any] | list | tuple = None,
        read_kwargs: dict | None = None,
        read_pd_kwargs: dict | None = None,
        df_type: Literal["pandas", "polars"] = "pandas",
        replace: bool = False,
        aws_conn_id: str | None = "aws_default",
        verify: bool | str | None = None,
        file_format: Literal["csv", "json", "parquet"] = "csv",
        max_rows_per_file: int = 0,
        df_kwargs: dict | None = None,
        pd_kwargs: dict | None = None,
        groupby_kwargs: dict | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.query = query
        self.s3_bucket = s3_bucket
        self.s3_key = s3_key
        self.sql_conn_id = sql_conn_id
        self.aws_conn_id = aws_conn_id
        self.verify = verify
        self.replace = replace
        self.parameters = parameters
        self.max_rows_per_file = max_rows_per_file
        self.groupby_kwargs = groupby_kwargs or {}
        self.sql_hook_params = sql_hook_params
        self.df_type = df_type

        if read_pd_kwargs is not None:
            warnings.warn(
                "The 'read_pd_kwargs' parameter is deprecated. Use 'read_kwargs' instead.",
                AirflowProviderDeprecationWarning,
                stacklevel=2,
            )
        self.read_kwargs = read_kwargs if read_kwargs is not None else read_pd_kwargs or {}

        if pd_kwargs is not None:
            warnings.warn(
                "The 'pd_kwargs' parameter is deprecated. Use 'df_kwargs' instead.",
                AirflowProviderDeprecationWarning,
                stacklevel=2,
            )

        self.df_kwargs = df_kwargs if df_kwargs is not None else pd_kwargs or {}

        if "path_or_buf" in self.df_kwargs:
            raise AirflowException("The argument path_or_buf is not allowed, please remove it")

        if self.max_rows_per_file and self.groupby_kwargs:
            raise AirflowException(
                "SqlToS3Operator arguments max_rows_per_file and groupby_kwargs "
                "can not be both specified. Please choose one."
            )

        try:
            self.file_format = FILE_FORMAT[file_format.upper()]
        except KeyError:
            raise AirflowException(f"The argument file_format doesn't support {file_format} value.")

    @staticmethod
    def _fix_dtypes(df: pd.DataFrame, file_format: FILE_FORMAT) -> None:
        """
        Mutate DataFrame to set dtypes for float columns containing NaN values.

        Set dtype of object to str to allow for downstream transformations.
        """
        try:
            import numpy as np
            import pandas as pd
        except ImportError as e:
            from airflow.exceptions import AirflowOptionalProviderFeatureException

            raise AirflowOptionalProviderFeatureException(e)

        for col in df:
            if df[col].dtype.name == "object" and file_format == FILE_FORMAT.PARQUET:
                # if the type wasn't identified or converted, change it to a string so if can still be
                # processed.
                df[col] = df[col].astype(str)

            if "float" in df[col].dtype.name and df[col].hasnans:
                # inspect values to determine if dtype of non-null values is int or float
                notna_series: Any = df[col].dropna().values
                if np.equal(notna_series, notna_series.astype(int)).all():
                    # set to dtype that retains integers and supports NaNs
                    # The type ignore can be removed here if https://github.com/numpy/numpy/pull/23690
                    # is merged and released as currently NumPy does not consider None as valid for x/y.
                    df[col] = np.where(df[col].isnull(), None, df[col])  # type: ignore[call-overload]
                    df[col] = df[col].astype(pd.Int64Dtype())
                elif np.isclose(notna_series, notna_series.astype(int)).all():
                    # set to float dtype that retains floats and supports NaNs
                    # The type ignore can be removed here if https://github.com/numpy/numpy/pull/23690
                    # is merged and released
                    df[col] = np.where(df[col].isnull(), None, df[col])  # type: ignore[call-overload]
                    df[col] = df[col].astype(pd.Float64Dtype())

    @staticmethod
    def _strip_suffixes(
        path: str,
    ) -> str:
        suffixes = [".json.gz", ".csv.gz", ".json", ".csv", ".parquet"]
        for suffix in sorted(suffixes, key=len, reverse=True):
            if path.endswith(suffix):
                return path[: -len(suffix)]
        return path

    def execute(self, context: Context) -> None:
        sql_hook = self._get_hook()
        s3_conn = S3Hook(aws_conn_id=self.aws_conn_id, verify=self.verify)
        data_df = sql_hook.get_df(
            sql=self.query, parameters=self.parameters, df_type=self.df_type, **self.read_kwargs
        )
        self.log.info("Data from SQL obtained")
        # Only apply dtype fixes to pandas DataFrames since Polars doesn't have the same NaN/None inconsistencies as panda
        if ("dtype_backend", "pyarrow") not in self.read_kwargs.items() and self.df_type == "pandas":
            self._fix_dtypes(data_df, self.file_format)  # type: ignore[arg-type]
        file_options = FILE_OPTIONS_MAP[self.file_format]

        for group_name, df in self._partition_dataframe(df=data_df):
            buf = io.BytesIO()
            self.log.info("Writing data to in-memory buffer")
            clean_key = self._strip_suffixes(self.s3_key)
            object_key = (
                f"{clean_key}_{group_name}{file_options.suffix}"
                if group_name
                else f"{clean_key}{file_options.suffix}"
            )

            if self.file_format != FILE_FORMAT.PARQUET and self.df_kwargs.get("compression") == "gzip":
                object_key += ".gz"
                df_kwargs = {k: v for k, v in self.df_kwargs.items() if k != "compression"}
                with gzip.GzipFile(fileobj=buf, mode="wb", filename=object_key) as gz:
                    getattr(df, file_options.function)(gz, **df_kwargs)
            else:
                if self.file_format == FILE_FORMAT.PARQUET:
                    getattr(df, file_options.function)(buf, **self.df_kwargs)
                else:
                    text_buf = io.TextIOWrapper(buf, encoding="utf-8", write_through=True)
                    getattr(df, file_options.function)(text_buf, **self.df_kwargs)
                    text_buf.flush()

            buf.seek(0)

            self.log.info("Uploading data to S3")
            s3_conn.load_file_obj(
                file_obj=buf, key=object_key, bucket_name=self.s3_bucket, replace=self.replace
            )

    def _partition_dataframe(
        self, df: pd.DataFrame | pl.DataFrame
    ) -> Iterable[tuple[str, pd.DataFrame | pl.DataFrame]]:
        """Partition dataframe using pandas or polars groupby() method."""
        try:
            import secrets
            import string

            import numpy as np
            import pandas as pd
            import polars as pl
        except ImportError:
            pass

        # if max_rows_per_file argument is specified, a temporary column with a random unusual name will be
        # added to the dataframe. This column is used to dispatch the dataframe into smaller ones using groupby()
        random_column_name = None
        if self.max_rows_per_file and not self.groupby_kwargs:
            random_column_name = "".join(secrets.choice(string.ascii_letters) for _ in range(20))
            self.groupby_kwargs = {"by": random_column_name}

        if random_column_name:
            if isinstance(df, pd.DataFrame):
                df[random_column_name] = np.arange(len(df)) // self.max_rows_per_file
            elif isinstance(df, pl.DataFrame):
                df = df.with_columns(
                    (pl.int_range(pl.len()) // self.max_rows_per_file).alias(random_column_name)
                )

        if not self.groupby_kwargs:
            yield "", df
            return

        if isinstance(df, pd.DataFrame):
            for group_label in (grouped_df := df.groupby(**self.groupby_kwargs)).groups:
                group_df = grouped_df.get_group(group_label)
                if random_column_name:
                    group_df = group_df.drop(random_column_name, axis=1, errors="ignore")
                yield (
                    cast("str", group_label[0] if isinstance(group_label, tuple) else group_label),
                    group_df.reset_index(drop=True),
                )
        elif isinstance(df, pl.DataFrame):
            for group_label, group_df_in in df.group_by(**self.groupby_kwargs):  # type: ignore[assignment]
                group_df2 = group_df_in.drop(random_column_name) if random_column_name else group_df_in
                yield (
                    cast("str", group_label[0] if isinstance(group_label, tuple) else group_label),
                    group_df2,
                )

    def _get_hook(self) -> DbApiHook:
        self.log.debug("Get connection for %s", self.sql_conn_id)
        conn = BaseHook.get_connection(self.sql_conn_id)
        hook = conn.get_hook(hook_params=self.sql_hook_params)
        if not callable(getattr(hook, "get_df", None)):
            raise AirflowException("This hook is not supported. The hook class must have get_df method.")
        return hook



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from airflow.configuration import conf
from airflow.providers.amazon.aws.hooks.ssm import SsmHook
from airflow.providers.amazon.aws.operators.base_aws import AwsBaseOperator
from airflow.providers.amazon.aws.triggers.ssm import SsmRunCommandTrigger
from airflow.providers.amazon.aws.utils import validate_execute_complete_event
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields

if TYPE_CHECKING:
    from airflow.utils.context import Context


class SsmRunCommandOperator(AwsBaseOperator[SsmHook]):
    """
    Executes the SSM Run Command to perform actions on managed instances.

    .. seealso::
        For more information on how to use this operator, take a look at the
        guide:
        :ref:`howto/operator:SsmRunCommandOperator`

    :param document_name: The name of the Amazon Web Services Systems Manager
        document (SSM document) to run.
    :param run_command_kwargs: Optional parameters to pass to the send_command
        API.

    :param wait_for_completion: Whether to wait for cluster to stop.
        (default: True)
    :param waiter_delay: Time in seconds to wait between status checks.
        (default: 120)
    :param waiter_max_attempts: Maximum number of attempts to check for job
        completion. (default: 75)
    :param deferrable: If True, the operator will wait asynchronously for the
        cluster to stop. This implies waiting for completion. This mode
        requires aiobotocore module to be installed. (default: False)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used.
        If running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default
        boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore
        client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = SsmHook
    template_fields: Sequence[str] = aws_template_fields(
        "document_name",
        "run_command_kwargs",
    )

    def __init__(
        self,
        *,
        document_name: str,
        run_command_kwargs: dict[str, Any] | None = None,
        wait_for_completion: bool = True,
        waiter_delay: int = 120,
        waiter_max_attempts: int = 75,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.wait_for_completion = wait_for_completion
        self.waiter_delay = waiter_delay
        self.waiter_max_attempts = waiter_max_attempts
        self.deferrable = deferrable

        self.document_name = document_name
        self.run_command_kwargs = run_command_kwargs or {}

    def execute_complete(self, context: Context, event: dict[str, Any] | None = None) -> str:
        event = validate_execute_complete_event(event)

        if event["status"] != "success":
            raise RuntimeError(f"Error while running run command: {event}")

        self.log.info("SSM run command `%s` completed.", event["command_id"])
        return event["command_id"]

    def execute(self, context: Context):
        response = self.hook.conn.send_command(
            DocumentName=self.document_name,
            **self.run_command_kwargs,
        )

        command_id = response["Command"]["CommandId"]
        task_description = f"SSM run command {command_id} to complete."

        if self.deferrable:
            self.log.info("Deferring for %s", task_description)
            self.defer(
                trigger=SsmRunCommandTrigger(
                    command_id=command_id,
                    waiter_delay=self.waiter_delay,
                    waiter_max_attempts=self.waiter_max_attempts,
                    aws_conn_id=self.aws_conn_id,
                    region_name=self.region_name,
                    verify=self.verify,
                    botocore_config=self.botocore_config,
                ),
                method_name="execute_complete",
            )

        elif self.wait_for_completion:
            self.log.info("Waiting for %s", task_description)
            waiter = self.hook.get_waiter("command_executed")

            instance_ids = response["Command"]["InstanceIds"]
            for instance_id in instance_ids:
                waiter.wait(
                    CommandId=command_id,
                    InstanceId=instance_id,
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )

        return command_id


class SsmGetCommandInvocationOperator(AwsBaseOperator[SsmHook]):
    """
    Retrieves the output and execution details of an SSM command invocation.

    This operator allows you to fetch the standard output, standard error,
    execution status, and other details from SSM commands. It can be used to
    retrieve output from commands executed by SsmRunCommandOperator in previous
    tasks, or from commands executed outside of Airflow entirely.

    The operator returns structured data including stdout, stderr, execution
    times, and status information for each instance that executed the command.

    .. seealso::
        For more information on how to use this operator, take a look at the
        guide:
        :ref:`howto/operator:SsmGetCommandInvocationOperator`

    :param command_id: The ID of the SSM command to retrieve output for.
    :param instance_id: The ID of the specific instance to retrieve output
        for. If not provided, retrieves output from all instances that
        executed the command.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used.
        If running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default
        boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore
        client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = SsmHook
    template_fields: Sequence[str] = aws_template_fields(
        "command_id",
        "instance_id",
    )

    def __init__(
        self,
        *,
        command_id: str,
        instance_id: str | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.command_id = command_id
        self.instance_id = instance_id

    def execute(self, context: Context) -> dict[str, Any]:
        """Execute the operator to retrieve command invocation output."""
        if self.instance_id:
            self.log.info(
                "Retrieving output for command %s on instance %s",
                self.command_id,
                self.instance_id,
            )
            invocations = [{"InstanceId": self.instance_id}]
        else:
            self.log.info("Retrieving output for command %s from all instances", self.command_id)
            response = self.hook.list_command_invocations(self.command_id)
            invocations = response.get("CommandInvocations", [])

        output_data: dict[str, Any] = {"command_id": self.command_id, "invocations": []}

        for invocation in invocations:
            instance_id = invocation["InstanceId"]
            try:
                invocation_details = self.hook.get_command_invocation(self.command_id, instance_id)
                output_data["invocations"].append(
                    {
                        "instance_id": instance_id,
                        "status": invocation_details.get("Status", ""),
                        "response_code": invocation_details.get("ResponseCode", ""),
                        "standard_output": invocation_details.get("StandardOutputContent", ""),
                        "standard_error": invocation_details.get("StandardErrorContent", ""),
                        "execution_start_time": invocation_details.get("ExecutionStartDateTime", ""),
                        "execution_end_time": invocation_details.get("ExecutionEndDateTime", ""),
                        "document_name": invocation_details.get("DocumentName", ""),
                        "comment": invocation_details.get("Comment", ""),
                    }
                )
            except Exception as e:
                self.log.warning("Failed to get output for instance %s: %s", instance_id, e)
                output_data["invocations"].append({"instance_id": instance_id, "error": str(e)})

        return output_data



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from typing import TYPE_CHECKING

from botocore.waiter import Waiter, WaiterModel, create_waiter_with_client

if TYPE_CHECKING:
    import boto3


class BaseBotoWaiter:
    """
    Used to create custom Boto3 Waiters.

    For more details, see airflow/providers/amazon/aws/waiters/README.md
    """

    def __init__(self, client: boto3.client, model_config: dict, deferrable: bool = False) -> None:
        self.model = WaiterModel(model_config)
        self.client = client
        self.deferrable = deferrable

    def _get_async_waiter_with_client(self, waiter_name: str):
        from aiobotocore.waiter import create_waiter_with_client as create_async_waiter_with_client

        return create_async_waiter_with_client(
            waiter_name=waiter_name, waiter_model=self.model, client=self.client
        )

    def waiter(self, waiter_name: str) -> Waiter:
        if self.deferrable:
            return self._get_async_waiter_with_client(waiter_name=waiter_name)

        return create_waiter_with_client(waiter_name=waiter_name, waiter_model=self.model, client=self.client)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Generator
from unittest import mock

import pytest

from airflow.providers.amazon.aws.hooks.ssm import SsmHook
from airflow.providers.amazon.aws.operators.ssm import SsmGetCommandInvocationOperator, SsmRunCommandOperator

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

COMMAND_ID = "test_command_id"
DOCUMENT_NAME = "test_ssm_custom_document"
INSTANCE_IDS = ["test_instance_id_1", "test_instance_id_2"]


class TestSsmRunCommandOperator:
    @pytest.fixture
    def mock_conn(self) -> Generator[SsmHook, None, None]:
        with mock.patch.object(SsmHook, "conn") as _conn:
            _conn.send_command.return_value = {
                "Command": {
                    "CommandId": COMMAND_ID,
                    "InstanceIds": INSTANCE_IDS,
                }
            }
            yield _conn

    def setup_method(self):
        self.operator = SsmRunCommandOperator(
            task_id="test_run_command_operator",
            document_name=DOCUMENT_NAME,
            run_command_kwargs={"InstanceIds": INSTANCE_IDS},
        )
        self.operator.defer = mock.MagicMock()

    @pytest.mark.parametrize(
        ("wait_for_completion", "deferrable"),
        [
            pytest.param(False, False, id="no_wait"),
            pytest.param(True, False, id="wait"),
            pytest.param(False, True, id="defer"),
        ],
    )
    @mock.patch.object(SsmHook, "get_waiter")
    def test_run_command_wait_combinations(self, mock_get_waiter, wait_for_completion, deferrable, mock_conn):
        self.operator.wait_for_completion = wait_for_completion
        self.operator.deferrable = deferrable

        command_id = self.operator.execute({})

        assert command_id == COMMAND_ID
        mock_conn.send_command.assert_called_once_with(DocumentName=DOCUMENT_NAME, InstanceIds=INSTANCE_IDS)
        assert mock_get_waiter.call_count == wait_for_completion
        assert self.operator.defer.call_count == deferrable

    def test_template_fields(self):
        validate_template_fields(self.operator)

    def test_deferrable_with_region(self, mock_conn):
        """Test that deferrable mode properly passes region and other AWS parameters to trigger."""
        self.operator.deferrable = True
        self.operator.region_name = "us-west-2"
        self.operator.verify = False
        self.operator.botocore_config = {"retries": {"max_attempts": 5}}

        command_id = self.operator.execute({})

        assert command_id == COMMAND_ID
        mock_conn.send_command.assert_called_once_with(DocumentName=DOCUMENT_NAME, InstanceIds=INSTANCE_IDS)

        # Verify defer was called with correct trigger parameters
        self.operator.defer.assert_called_once()
        call_args = self.operator.defer.call_args
        trigger = call_args[1]["trigger"]  # Get the trigger from kwargs

        # Verify the trigger has the correct parameters
        assert trigger.command_id == COMMAND_ID
        assert trigger.region_name == "us-west-2"
        assert trigger.verify is False
        assert trigger.botocore_config == {"retries": {"max_attempts": 5}}
        assert trigger.aws_conn_id == self.operator.aws_conn_id


class TestSsmGetCommandInvocationOperator:
    @pytest.fixture
    def mock_hook(self) -> Generator[mock.MagicMock, None, None]:
        with mock.patch.object(SsmGetCommandInvocationOperator, "hook") as _hook:
            yield _hook

    def setup_method(self):
        self.command_id = "test-command-id-123"
        self.instance_id = "i-1234567890abcdef0"
        self.operator = SsmGetCommandInvocationOperator(
            task_id="test_get_command_invocation",
            command_id=self.command_id,
            instance_id=self.instance_id,
        )

    def test_execute_with_specific_instance(self, mock_hook):
        # Mock response for specific instance
        mock_invocation_details = {
            "Status": "Success",
            "ResponseCode": 0,
            "StandardOutputContent": "Hello World",
            "StandardErrorContent": "",
            "ExecutionStartDateTime": "2023-01-01T12:00:00Z",
            "ExecutionEndDateTime": "2023-01-01T12:00:05Z",
            "DocumentName": "AWS-RunShellScript",
            "Comment": "Test command",
        }
        mock_hook.get_command_invocation.return_value = mock_invocation_details

        result = self.operator.execute({})

        # Verify hook was called correctly
        mock_hook.get_command_invocation.assert_called_once_with(self.command_id, self.instance_id)

        # Verify returned data structure - should use standardized format with invocations array
        expected_result = {
            "command_id": self.command_id,
            "invocations": [
                {
                    "instance_id": self.instance_id,
                    "status": "Success",
                    "response_code": 0,
                    "standard_output": "Hello World",
                    "standard_error": "",
                    "execution_start_time": "2023-01-01T12:00:00Z",
                    "execution_end_time": "2023-01-01T12:00:05Z",
                    "document_name": "AWS-RunShellScript",
                    "comment": "Test command",
                }
            ],
        }
        assert result == expected_result

    def test_execute_all_instances(self, mock_hook):
        # Setup operator without instance_id to get all instances
        operator = SsmGetCommandInvocationOperator(
            task_id="test_get_all_invocations",
            command_id=self.command_id,
        )

        # Mock list_command_invocations response
        mock_invocations = [
            {"InstanceId": "i-111"},
            {"InstanceId": "i-222"},
        ]
        mock_hook.list_command_invocations.return_value = {"CommandInvocations": mock_invocations}

        # Mock get_command_invocation responses
        mock_invocation_details_1 = {
            "Status": "Success",
            "ResponseCode": 0,
            "StandardOutputContent": "Output 1",
            "StandardErrorContent": "",
            "ExecutionStartDateTime": "2023-01-01T12:00:00Z",
            "ExecutionEndDateTime": "2023-01-01T12:00:05Z",
            "DocumentName": "AWS-RunShellScript",
            "Comment": "",
        }
        mock_invocation_details_2 = {
            "Status": "Failed",
            "ResponseCode": 1,
            "StandardOutputContent": "",
            "StandardErrorContent": "Error occurred",
            "ExecutionStartDateTime": "2023-01-01T12:00:00Z",
            "ExecutionEndDateTime": "2023-01-01T12:00:10Z",
            "DocumentName": "AWS-RunShellScript",
            "Comment": "",
        }

        mock_hook.get_command_invocation.side_effect = [
            mock_invocation_details_1,
            mock_invocation_details_2,
        ]

        result = operator.execute({})

        # Verify hook calls
        mock_hook.list_command_invocations.assert_called_once_with(self.command_id)
        assert mock_hook.get_command_invocation.call_count == 2
        mock_hook.get_command_invocation.assert_any_call(self.command_id, "i-111")
        mock_hook.get_command_invocation.assert_any_call(self.command_id, "i-222")

        # Verify returned data structure
        expected_result = {
            "command_id": self.command_id,
            "invocations": [
                {
                    "instance_id": "i-111",
                    "status": "Success",
                    "response_code": 0,
                    "standard_output": "Output 1",
                    "standard_error": "",
                    "execution_start_time": "2023-01-01T12:00:00Z",
                    "execution_end_time": "2023-01-01T12:00:05Z",
                    "document_name": "AWS-RunShellScript",
                    "comment": "",
                },
                {
                    "instance_id": "i-222",
                    "status": "Failed",
                    "response_code": 1,
                    "standard_output": "",
                    "standard_error": "Error occurred",
                    "execution_start_time": "2023-01-01T12:00:00Z",
                    "execution_end_time": "2023-01-01T12:00:10Z",
                    "document_name": "AWS-RunShellScript",
                    "comment": "",
                },
            ],
        }
        assert result == expected_result

    def test_execute_all_instances_with_error(self, mock_hook):
        # Setup operator without instance_id
        operator = SsmGetCommandInvocationOperator(
            task_id="test_get_all_with_error",
            command_id=self.command_id,
        )

        # Mock list_command_invocations response
        mock_invocations = [{"InstanceId": "i-111"}]
        mock_hook.list_command_invocations.return_value = {"CommandInvocations": mock_invocations}

        # Mock get_command_invocation to raise an exception
        mock_hook.get_command_invocation.side_effect = Exception("API Error")

        result = operator.execute({})

        # Verify error handling
        expected_result = {
            "command_id": self.command_id,
            "invocations": [{"instance_id": "i-111", "error": "API Error"}],
        }
        assert result == expected_result

    def test_template_fields(self):
        validate_template_fields(self.operator)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

from airflow.providers.amazon.aws.hooks.sqs import SqsHook
from airflow.providers.amazon.aws.operators.sqs import SqsPublishOperator
from airflow.providers.amazon.aws.sensors.sqs import SqsSensor

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task
else:
    # Airflow 2.10 compat
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]
try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import SystemTestContextBuilder

sys_test_context_task = SystemTestContextBuilder().build()

DAG_ID = "example_sqs"


@task
def create_queue(queue_name) -> str:
    return SqsHook().create_queue(queue_name=queue_name)["QueueUrl"]


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_queue(queue_url):
    SqsHook().conn.delete_queue(QueueUrl=queue_url)


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    tags=["example"],
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context["ENV_ID"]

    sns_queue_name = f"{env_id}-example-queue"

    sqs_queue = create_queue(sns_queue_name)

    # [START howto_operator_sqs]
    publish_to_queue_1 = SqsPublishOperator(
        task_id="publish_to_queue_1",
        sqs_queue=sqs_queue,
        message_content="{{ task_instance }}",
    )
    publish_to_queue_2 = SqsPublishOperator(
        task_id="publish_to_queue_2",
        sqs_queue=sqs_queue,
        message_content="{{ task_instance }}",
    )
    # [END howto_operator_sqs]

    # [START howto_sensor_sqs]
    read_from_queue = SqsSensor(
        task_id="read_from_queue",
        sqs_queue=sqs_queue,
    )
    # Retrieve multiple batches of messages from SQS.
    # The SQS API only returns a maximum of 10 messages per poll.
    read_from_queue_in_batch = SqsSensor(
        task_id="read_from_queue_in_batch",
        sqs_queue=sqs_queue,
        # Get maximum 10 messages each poll
        max_messages=10,
        # Combine 3 polls before returning results
        num_batches=3,
    )
    # [END howto_sensor_sqs]

    chain(
        # TEST SETUP
        test_context,
        sqs_queue,
        # TEST BODY
        publish_to_queue_1,
        read_from_queue,
        publish_to_queue_2,
        read_from_queue_in_batch,
        # TEST TEARDOWN
        delete_queue(sqs_queue),
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime
import logging
import textwrap
import time

import boto3

from airflow.providers.amazon.aws.operators.ec2 import EC2CreateInstanceOperator, EC2TerminateInstanceOperator
from airflow.providers.amazon.aws.operators.ssm import SsmGetCommandInvocationOperator, SsmRunCommandOperator
from airflow.providers.amazon.aws.sensors.ssm import SsmRunCommandCompletedSensor
from airflow.sdk import DAG, chain, task

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder, get_role_name
from system.amazon.aws.utils.ec2 import get_latest_ami_id

DAG_ID = "example_ssm"

ROLE_ARN_KEY = "ROLE_ARN"
sys_test_context_task = SystemTestContextBuilder().add_variable(ROLE_ARN_KEY).build()

USER_DATA = textwrap.dedent("""\
    #!/bin/bash
    set -e

    # Update the system
    if command -v yum &> /dev/null; then
        PACKAGE_MANAGER="yum"
    elif command -v dnf &> /dev/null; then
        PACKAGE_MANAGER="dnf"
    else
        echo "No suitable package manager found"
        exit 1
    fi

    # Install SSM agent if it's not installed
    if ! command -v amazon-ssm-agent &> /dev/null; then
        echo "Installing SSM agent..."
        $PACKAGE_MANAGER install -y amazon-ssm-agent
    else
        echo "SSM agent already installed"
    fi

    echo "Enabling and starting SSM agent..."
    systemctl enable amazon-ssm-agent
    systemctl start amazon-ssm-agent

    shutdown -h +15

    echo "=== Finished user-data script ==="
""")

log = logging.getLogger(__name__)


@task
def create_instance_profile(role_name: str, instance_profile_name: str):
    client = boto3.client("iam")
    client.create_instance_profile(InstanceProfileName=instance_profile_name)
    client.add_role_to_instance_profile(InstanceProfileName=instance_profile_name, RoleName=role_name)


@task
def await_instance_profile_exists(instance_profile_name):
    client = boto3.client("iam")
    client.get_waiter("instance_profile_exists").wait(InstanceProfileName=instance_profile_name)


@task
def delete_instance_profile(instance_profile_name, role_name):
    client = boto3.client("iam")

    try:
        client.remove_role_from_instance_profile(
            InstanceProfileName=instance_profile_name, RoleName=role_name
        )
    except client.exceptions.NoSuchEntityException:
        log.info("Role %s not attached to %s or already removed.", role_name, instance_profile_name)

    try:
        client.delete_instance_profile(InstanceProfileName=instance_profile_name)
    except client.exceptions.NoSuchEntityException:
        log.info("Instance profile %s already deleted.", instance_profile_name)


@task
def extract_instance_id(instance_ids: list) -> str:
    return instance_ids[0]


@task
def build_run_command_kwargs(instance_id: str):
    return {
        "InstanceIds": [instance_id],
        "Parameters": {"commands": ["touch /tmp/ssm_test_passed"]},
    }


@task
def wait_until_ssm_ready(instance_id: str, max_attempts: int = 10, delay_seconds: int = 15):
    """
    Waits for an EC2 instance to register with AWS Systems Manager (SSM).

    This may take over a minute even after the instance is running.
    Raises an exception if the instance is not ready after max_attempts.
    """
    ssm = boto3.client("ssm")

    for _ in range(max_attempts):
        response = ssm.describe_instance_information(
            Filters=[{"Key": "InstanceIds", "Values": [instance_id]}]
        )

        if (
            response.get("InstanceInformationList")
            and response["InstanceInformationList"][0]["PingStatus"] == "Online"
        ):
            return

        time.sleep(delay_seconds)

    raise Exception(f"Instance {instance_id} not ready in SSM after {max_attempts} attempts.")


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime.datetime(2021, 1, 1),
    tags=["example"],
    catchup=False,
) as dag:
    # Create EC2 instance with SSM agent
    test_context = sys_test_context_task()
    env_id = test_context[ENV_ID_KEY]
    instance_name = f"{env_id}-instance"
    image_id = get_latest_ami_id()
    role_name = get_role_name(test_context[ROLE_ARN_KEY])
    instance_profile_name = f"{env_id}-ssm-instance-profile"

    config = {
        "InstanceType": "t4g.micro",
        "IamInstanceProfile": {"Name": instance_profile_name},
        # Optional: Tags for identifying test resources in the AWS console
        "TagSpecifications": [
            {"ResourceType": "instance", "Tags": [{"Key": "Name", "Value": instance_name}]}
        ],
        "UserData": USER_DATA,
        # Use IMDSv2 for greater security, see the following doc for more details:
        # https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/configuring-instance-metadata-service.html
        "MetadataOptions": {"HttpEndpoint": "enabled", "HttpTokens": "required"},
        "BlockDeviceMappings": [
            {"DeviceName": "/dev/xvda", "Ebs": {"Encrypted": True, "DeleteOnTermination": True}}
        ],
        "InstanceInitiatedShutdownBehavior": "terminate",
    }

    create_instance = EC2CreateInstanceOperator(
        task_id="create_instance",
        image_id=image_id,
        max_count=1,
        min_count=1,
        config=config,
        wait_for_completion=True,
        retries=5,
        retry_delay=datetime.timedelta(seconds=15),
    )

    instance_id = extract_instance_id(create_instance.output)

    run_command_kwargs = build_run_command_kwargs(instance_id)

    # [START howto_operator_run_command]
    run_command = SsmRunCommandOperator(
        task_id="run_command",
        document_name="AWS-RunShellScript",
        run_command_kwargs=run_command_kwargs,
        wait_for_completion=False,
    )
    # [END howto_operator_run_command]

    # [START howto_sensor_run_command]
    await_run_command = SsmRunCommandCompletedSensor(
        task_id="await_run_command", command_id="{{ ti.xcom_pull(task_ids='run_command') }}"
    )
    # [END howto_sensor_run_command]

    # [START howto_operator_get_command_invocation]
    get_command_output = SsmGetCommandInvocationOperator(
        task_id="get_command_output",
        command_id="{{ ti.xcom_pull(task_ids='run_command') }}",
        instance_id=instance_id,
    )
    # [END howto_operator_get_command_invocation]

    delete_instance = EC2TerminateInstanceOperator(
        task_id="terminate_instance",
        trigger_rule=TriggerRule.ALL_DONE,
        instance_ids=instance_id,
    )

    chain(
        # TEST SETUP
        test_context,
        image_id,
        role_name,
        create_instance_profile(role_name, instance_profile_name),
        await_instance_profile_exists(instance_profile_name),
        create_instance,
        instance_id,
        run_command_kwargs,
        wait_until_ssm_ready(instance_id),
        # TEST BODY
        run_command,
        await_run_command,
        get_command_output,
        # TEST TEARDOWN
        delete_instance,
        delete_instance_profile(instance_profile_name, role_name),
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)
